/**********************************************************************
 * Ngaro Virtual Machine
 * Copyright (C) 2008 - 2011, Charles Childers
 **********************************************************************/

function Stack(size)
{
  this.sp = 0;
  this.data = new Array(size);
  this.push = function(n) { this.sp++; this.data[this.sp] = n; }
  this.pop  = function()  { var n = this.data[this.sp]; this.sp--; return n; }
  this.depth = function() { return this.sp; }
  this.tos = function() { return this.data[this.sp]; }
  this.nos = function() { return this.data[this.sp - 1]; }
  this.dup = function() { var n = this.tos(); this.push(n); }
  this.drop = function() { this.sp--; }
  this.swap = function() { var a = this.pop(); var b = this.pop(); this.push(a); this.push(b); }
  this.inc  = function() { this.data[this.sp]++; }
  this.dec  = function() { this.data[this.sp]--; }
  this.reset = function() { this.sp = 0; }
}


/**********************************************************************
 * Symbolic constants for each instruction.
 **********************************************************************/
function Opcodes()
{
  this.NOP = 0;       this.LIT = 1;         this.DUP = 2;
  this.DROP = 3;      this.SWAP = 4;        this.PUSH = 5;
  this.POP = 6;       this.LOOP = 7;        this.JUMP = 8;
  this.RETURN = 9;    this.GT_JUMP = 10;    this.LT_JUMP = 11;
  this.NE_JUMP = 12;  this.EQ_JUMP = 13;    this.FETCH = 14;
  this.STORE = 15;    this.ADD = 16;        this.SUB = 17;
  this.MUL = 18;      this.DIVMOD = 19;     this.AND = 20;
  this.OR = 21;       this.XOR = 22;        this.SHL = 23;
  this.SHR = 24;      this.ZERO_EXIT = 25;  this.INC = 26;
  this.DEC = 27;      this.IN = 28;         this.OUT = 29;
  this.WAIT = 30;
}


/**********************************************************************
 * Some constants useful to us for dealing with the VM settings.
 *
 * If you have performance issues, try modifying CYLES_PER
 **********************************************************************/
  var   IMAGE_SIZE    = 256000;         /* Amount of simulated RAM    */
  var   DATA_DEPTH    =    128;         /* Depth of data stack        */
  var   ADDRESS_DEPTH =   1000;         /* Depth of the stacks        */
  var   CYCLES_PER    =   4000;         /* Instructions to run per    */
                                        /* clock cycle                */
  var   TERM_WIDTH    =     80;         /* Width of virtual terminal  */
  var   TERM_HEIGHT   =     22;         /* Height of virtual terminal */
  var   FB_WIDTH      =      0;         /* Canvas Width               */
  var   FB_HEIGHT     =      0;         /* Canvas Height              */
  var   FB_EXISTS     =      0;         /* Is Canvas Present?         */


/**********************************************************************
 * Internal registers, flags, and variables
 **********************************************************************/
  var ip = 0;
  var run = 0;
  var data    = new Stack(DATA_DEPTH);
  var address = new Stack(ADDRESS_DEPTH);
  var ports   = new Array(64);
  var portHandlers = new Array(64);
  var image   = new Array(IMAGE_SIZE);
  var interval;
  var devOutput = "";
  var output = document.getElementById("output");
  var lastKey = " ";
  var width = 0;
  var vm = new Opcodes();
  var tib = " ";


/**********************************************************************
 * getImage()
 * Load a saved image
 **********************************************************************/
function getImage()
{
  clearDisplay();
  stopVM();
  ip = 0;
  try
  {
    image = localStorage['retroImage'].split(';').map(function(x){return parseInt(x)});
  }
  catch (e)
  {
    alert("Sorry, unable to find a saved image");
  }
  data.reset();
  address.reset();
  startVM();
}


/**********************************************************************
 * saveImage()
 * Save the current image
 **********************************************************************/
function saveImage()
{
  stopVM();
  try
  {
    localStorage.setItem("retroImage", image.join(";"));
  }
  catch (e)
  {
    if (e == QUOTA_EXCEEDED_ERR)
      alert("Quota exceeded!");
  }
  startVM();
}


/**********************************************************************
 * pristine()
 * Load the default image
 **********************************************************************/
function pristine()
{
  clearDisplay();
  stopVM();
  loadImage();
  ip = 0;
  data.reset();
  address.reset();
  startVM();
}


/**********************************************************************
 * initVM()
 * Initialize the Ngaro VM
 **********************************************************************/
function initVM()
{
  ip  = 0;
  ports[0] = 0;
  width = 0;
  var i = 0;
  while (i < 64)
  {
    ports[i] = 0;
    i++;
  }
  data.reset();
  address.reset();
}


/**********************************************************************
 * This is a nice little hack to read key presses and store them
 * somewhere for later use. It's critical for the console emulation.
 **********************************************************************/
function readKeyboard(e)
{
  var uni = e.keyCode ? e.keyCode : e.charCode;
  lastKey = uni;
  if (uni == 8)
    return false;
}


/**********************************************************************
 * handleDevices()
 * This handles the simulated hardware devices.
 **********************************************************************/
portHandlers[2] = function()
{
  var ch = String.fromCharCode(data.tos());

  /* Remap select characters to HTML */
  switch (data.tos())
  {
    case 10: ch = "<br>\n"; width = 0; break;
    case 32: ch = "&nbsp;"; break;
    case 38: ch = "&amp;";  break;
    case 60: ch = "&lt;";   break;
    case 62: ch = "&gt;";   break;
  }

  /* Display the character */
  if (data.tos() < 0)
    clearDisplay();
  else
  {
    devOutput += ch;
    width++;
    if (width > TERM_WIDTH)
    {
      width = 0;
      devOutput += "<br>";
    }
  }

  if (data.tos() == 8)
    devOutput = devOutput.substr(0, devOutput.length - 2);

  data.drop();

  ports[2] = 0;
}

portHandlers[4] = function()
{
  ports[4] = 0;
  saveImage();
}

portHandlers[5] = function()
{
  /* Capabilities */
  if (ports[5] == -1)
    ports[5] = IMAGE_SIZE;
  if (ports[5] == -2)
    ports[5] = FB_EXISTS;
  if (ports[5] == -3)
    ports[5] = FB_WIDTH;
  if (ports[5] == -4)
    ports[5] = FB_HEIGHT;
  if (ports[5] == -5)
    ports[5] = data.depth();
  if (ports[5] == -6)
    ports[5] = address.depth();
  if (ports[5] == -7)
    ports[5] = -1;
  if (ports[5] == -8)
  {
    var foo = new Date;
    var unixtime_ms = foo.getTime();
    var unixtime = parseInt(unixtime_ms / 1000);
    ports[5] = unixtime;
  }
  if (ports[5] == -9)
    ports[5] = 0;
  if (ports[5] == -11)
    ports[5] = TERM_WIDTH;
  if (ports[5] == -12)
    ports[5] = TERM_HEIGHT;
}

function handleDevices()
{
  if (ports[0] != 0)
    return;

  ports[0] = 1;

  /* Input */
  if (ports[1] == 1)
  {
    ports[1] = tib.charCodeAt(0);
    tib = tib.substr(1, tib.length - 1);
    lastKey = 0;
    return;
  }

  if (ports[2] == 1)
  {
    if (typeof portHandlers[2] == 'function')
      portHandlers[2]();
    return;
  }

  for (var a = 4; a < 64; a++)
  {
    if (ports[a] != 0)
    {
      if (typeof portHandlers[a] == 'function')
      {
        portHandlers[a]();
        return;
      }
    }
  }
}



/**********************************************************************
 * clearDisplay()
 * This clears the display.
 **********************************************************************/
function clearDisplay()
{
  devOutput = "";
  output.innerHTML = devOutput;
  width = 0;
  if (FB_EXISTS == -1)
    rxClearCanvas();
}



/**********************************************************************
 * processOpcode()
 * This is the main piece of code in Ngaro. It looks up each opcode and
 * carries out the proper action. This is mostly a straight port of the
 * C implementation, so some optimization opportunities will probably
 * exist.
 **********************************************************************/
function processOpcode()
{
  var x, y, z, op;
  op = image[ip];
  switch(op)
  {
    case vm.NOP:
      break;
    case vm.LIT:
      ip++; data.push(image[ip]);
      break;
    case vm.DUP:
      data.dup();
      break;
    case vm.DROP:
      data.drop();
      break;
    case vm.SWAP:
      data.swap();
      break;
    case vm.PUSH:
      address.push(data.pop());
      break;
    case vm.POP:
      data.push(address.pop())
      break;
    case vm.LOOP:
      data.dec();
      if (data.tos() != 0)
      {
        ip++;
        ip = image[ip] - 1;
      }
      else
      {
        ip++;
        data.drop();
      }
      break;
    case vm.JUMP:
      ip++;
      ip = image[ip] - 1;
      if (image[ip + 1] == 0) ip++;
      if (image[ip + 1] == 0) ip++;
      break;
    case vm.RETURN:
      ip = address.pop();
      break;
    case vm.GT_JUMP:
      ip++
      if (data.nos() > data.tos())
        ip = image[ip] - 1;
      data.drop();
      data.drop();
      break;
    case vm.LT_JUMP:
      ip++
      if (data.nos() < data.tos())
        ip = image[ip] - 1;
      data.drop();
      data.drop();
      break;
    case vm.NE_JUMP:
      ip++
      if (data.nos() != data.tos())
        ip = image[ip] - 1;
      data.drop();
      data.drop();
      break;
    case vm.EQ_JUMP:
      ip++
      if (data.nos() == data.tos())
        ip = image[ip] - 1;
      data.drop();
      data.drop();
      break;
    case vm.FETCH:
      x = data.pop();
      data.push(image[x]);
      break;
    case vm.STORE:
      image[data.tos()] = data.nos();
      data.drop();
      data.drop();
      break;
    case vm.ADD:
      x = data.pop();
      y = data.pop();
      data.push(x + y);
      break;
    case vm.SUB:
      x = data.pop();
      y = data.pop();
      data.push(y - x);
      break;
    case vm.MUL:
      x = data.pop();
      y = data.pop();
      data.push(y * x);
      break;
    case vm.DIVMOD:
      b = data.pop();
      a = data.pop();
      if (b == 0)
      {
        devOutput += "<br>ERROR: Divide by Zero; restarting<br>";
        ip = 0;
        data.sp = 0;
        address.sp = 0;
      }
      else
      {
        x = Math.abs(b);
        y = Math.abs(a);
        q = Math.floor(y / x);
        r = y % x;
        if (a < 0 && b < 0)
          r = r * -1;
        if (a > 0 && b < 0)
          q = q * -1;
        if (a < 0 && b > 0)
        {
          r = r * -1;
          q = q * -1;
        }
        data.push(r);
        data.push(q);
      }
      break;
    case vm.AND:
      x = data.pop();
      y = data.pop();
      data.push(x & y);
      break;
    case vm.OR:
      x = data.pop();
      y = data.pop();
      data.push(x | y);
      break;
    case vm.XOR:
      x = data.pop();
      y = data.pop();
      data.push(x ^ y);
      break;
    case vm.SHL:
      x = data.pop();
      y = data.pop();
      data.push(y << x);
      break;
    case vm.SHR:
      x = data.pop();
      y = data.pop();
      data.push(y >>= x);
      break;
    case vm.ZERO_EXIT:
      if (data.tos() == 0)
      {
        data.drop();
        ip = address.pop();
      }
      break;
    case vm.INC:
      data.inc();
      break;
    case vm.DEC:
      data.dec();
      break;
    case vm.IN:
      x = data.pop();
      data.push(ports[x]);
      ports[x] = 0;
      break;
    case vm.OUT:
      x = data.pop();
      y = data.pop();
      ports[x] = y;
      break;
    case vm.WAIT:
      handleDevices();
      break;
    default:
      address.push(ip);
      ip = image[ip] - 1;
      if (image[ip + 1] == 0) ip++;
      if (image[ip + 1] == 0) ip++;
      break;
  }
}




/**********************************************************************
 * startVM()
 **********************************************************************/
function startVM()
{
  interval = setInterval("processImage()", 10);
  run = 1;
}

function stopVM()
{
  clearInterval(interval);
  interval = null;
  run = 0;
}


/**********************************************************************
 * checkStack()
 **********************************************************************/
function checkStack()
{
  if (data.depth() < 0 || address.depth() < 0)
    devOutput += "<br>ERROR: Stack Underflow, restarting...<br>";
  if (data.depth() > DATA_DEPTH || address.depth() > DATA_DEPTH)
    devOutput += "<br>ERROR: Stack Overflow, restarting...<br>";
  if (data.depth() < 0 || address.depth() < 0 || data.depth() > DATA_DEPTH || address.depth() > ADDRESS_DEPTH)
  {
    ip = 0;
    data.sp = 0;
    address.sp = 0;
  }
}


/**********************************************************************
 * processImage()
 * This runs through the image, calling processOpcode() for each
 * instruction.
 *
 * For performance reasons up to CYCLES_PER instructions will be
 * executed by this code per call.
 **********************************************************************/
function processImage()
{
  var a;

  for (a = CYCLES_PER; a > 0 && run == 1 && tib != ""; a--)
  {
    processOpcode();
    checkStack();
    ip++;
  }

  /* Update the display */
  if (ports[3] == 0)
  {
    ports[3] = 1;
    output.innerHTML = devOutput;
    output.scrollTop = output.scrollHeight;
  }
}


function rxProcessInput()
{
  tib = document.getElementById('tib').value + "  ";
  document.getElementById('tib').value = "";
}
